@@ -1,35 +1,35 @@
-# $Id: Piece.pm,v 1.2 2001/10/22 12:37:11 matt Exp $
-
 package Time::Piece;
 
 use strict;
-use vars qw($VERSION @ISA @EXPORT %EXPORT_TAGS);
 
 require Exporter;
 require DynaLoader;
 use Time::Seconds;
 use Carp;
-use UNIVERSAL qw(isa);
+use Time::Local;
 
-@ISA = qw(Exporter DynaLoader);
+our @ISA = qw(Exporter DynaLoader);
 
-@EXPORT = qw(
+our @EXPORT = qw(
     localtime
     gmtime
 );
 
-%EXPORT_TAGS = (
+our %EXPORT_TAGS = (
     ':override' => 'internal',
     );
 
-$VERSION = '1.00';
+our $VERSION = '1.25_01';
 
 bootstrap Time::Piece $VERSION;
 
 my $DATE_SEP = '-';
 my $TIME_SEP = ':';
 my @MON_LIST = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
+my @FULLMON_LIST = qw(January February March April May June July
+                      August September October November December);
 my @DAY_LIST = qw(Sun Mon Tue Wed Thu Fri Sat);
+my @FULLDAY_LIST = qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday);
 
 use constant 'c_sec' => 0;
 use constant 'c_min' => 1;
@@ -44,15 +44,19 @@ use constant 'c_epoch' => 9;
 use constant 'c_islocal' => 10;
 
 sub localtime {
-    my $time = shift;
+    unshift @_, __PACKAGE__ unless eval { $_[0]->isa('Time::Piece') };
+    my $class = shift;
+    my $time  = shift;
     $time = time if (!defined $time);
-    _mktime($time, 1);
+    $class->_mktime($time, 1);
 }
 
 sub gmtime {
-    my $time = shift;
+    unshift @_, __PACKAGE__ unless eval { $_[0]->isa('Time::Piece') };
+    my $class = shift;
+    my $time  = shift;
     $time = time if (!defined $time);
-    _mktime($time, 0);
+    $class->_mktime($time, 0);
 }
 
 sub new {
@@ -62,16 +66,16 @@ sub new {
     my $self;
     
     if (defined($time)) {
-        $self = &localtime($time);
+        $self = $class->localtime($time);
     }
     elsif (ref($class) && $class->isa(__PACKAGE__)) {
-        $self = _mktime($class->[c_epoch], $class->[c_islocal]);
+        $self = $class->_mktime($class->epoch, $class->[c_islocal]);
     }
     else {
-        $self = &localtime();
+        $self = $class->localtime();
     }
     
-    return bless $self, $class;
+    return bless $self, ref($class) || $class;
 }
 
 sub parse {
@@ -89,12 +93,38 @@ sub parse {
 }
 
 sub _mktime {
-    my ($time, $islocal) = @_;
+    my ($class, $time, $islocal) = @_;
+    $class = eval { (ref $class) && (ref $class)->isa('Time::Piece') }
+           ? ref $class
+           : $class;
+    if (ref($time)) {
+        $time->[c_epoch] = undef;
+        return wantarray ? @$time : bless [@$time[0..9], $islocal], $class;
+    }
+    _tzset();
     my @time = $islocal ?
             CORE::localtime($time)
                 :
             CORE::gmtime($time);
-    wantarray ? @time : bless [@time, $time, $islocal], 'Time::Piece';
+    wantarray ? @time : bless [@time, $time, $islocal], $class;
+}
+
+my %_special_exports = (
+  localtime => sub { my $c = $_[0]; sub { $c->localtime(@_) } },
+  gmtime    => sub { my $c = $_[0]; sub { $c->gmtime(@_)    } },
+);
+
+sub export {
+  my ($class, $to, @methods) = @_;
+  for my $method (@methods) {
+    if (exists $_special_exports{$method}) {
+      no strict 'refs';
+      no warnings 'redefine';
+      *{$to . "::$method"} = $_special_exports{$method}->($class);
+    } else {
+      $class->SUPER::export($to, $method);
+    }
+  }
 }
 
 sub import {
@@ -124,7 +154,7 @@ sub min {
     $time->[c_min];
 }
 
-*minute = \&minute;
+*minute = \&min;
 
 sub hour {
     my $time = shift;
@@ -157,12 +187,25 @@ sub month {
         return $MON_LIST[$time->[c_mon]];
     }
     else {
-        return $time->strftime('%B');
+        return $time->strftime('%b');
     }
 }
 
 *monname = \&month;
 
+sub fullmonth {
+    my $time = shift;
+    if (@_) {
+        return $_[$time->[c_mon]];
+    }
+    elsif (@FULLMON_LIST) {
+        return $FULLMON_LIST[$time->[c_mon]];
+    }
+    else {
+        return $time->strftime('%B');
+    }
+}
+
 sub year {
     my $time = shift;
     $time->[c_year] + 1900;
@@ -175,7 +218,8 @@ sub _year {
 
 sub yy {
     my $time = shift;
-    $time->[c_year] % 100;
+    my $res = $time->[c_year] % 100;
+    return $res > 9 ? $res : "0$res";
 }
 
 sub wday {
@@ -199,12 +243,25 @@ sub wdayname {
         return $DAY_LIST[$time->[c_wday]];
     }
     else {
-        return $time->strftime('%A');
+        return $time->strftime('%a');
     }
 }
 
 *day = \&wdayname;
 
+sub fullday {
+    my $time = shift;
+    if (@_) {
+        return $_[$time->[c_wday]];
+    }
+    elsif (@FULLDAY_LIST) {
+        return $FULLDAY_LIST[$time->[c_wday]];
+    }
+    else {
+        return $time->strftime('%A');
+    }
+}
+
 sub yday {
     my $time = shift;
     $time->[c_yday];
@@ -222,29 +279,24 @@ sub isdst {
 # Thanks to Tony Olekshy <olekshy@cs.ualberta.ca> for this algorithm
 sub tzoffset {
     my $time = shift;
+    
+    return Time::Seconds->new(0) unless $time->[c_islocal];
 
-    my $epoch = $time->[c_epoch];
+    my $epoch = $time->epoch;
 
-    my $j = sub { # Tweaked Julian day number algorithm.
+    my $j = sub {
 
         my ($s,$n,$h,$d,$m,$y) = @_; $m += 1; $y += 1900;
 
-        # Standard Julian day number algorithm without constant.
-        #
-        my $y1 = $m > 2 ? $y : $y - 1;
+        $time->_jd($y, $m, $d, $h, $n, $s);
 
-        my $m1 = $m > 2 ? $m + 1 : $m + 13;
-
-        my $day = int(365.25 * $y1) + int(30.6001 * $m1) + $d;
-
-        # Modify to include hours/mins/secs in floating portion.
-        #
-        return $day + ($h + ($n + $s / 60) / 60) / 24;
     };
 
     # Compute floating offset in hours.
     #
-    my $delta = 24 * (&$j(CORE::localtime $epoch) - &$j(CORE::gmtime $epoch));
+    # Note use of crt methods so the tz is properly set...
+    # See: http://perlmonks.org/?node_id=820347
+    my $delta = 24 * ($j->(_crt_localtime($epoch)) - $j->(_crt_gmtime($epoch)));
 
     # Return value in seconds rounded to nearest minute.
     return Time::Seconds->new( int($delta * 60 + ($delta >= 0 ? 0.5 : -0.5)) * 60 );
@@ -252,7 +304,17 @@ sub tzoffset {
 
 sub epoch {
     my $time = shift;
-    $time->[c_epoch];
+    if (defined($time->[c_epoch])) {
+        return $time->[c_epoch];
+    }
+    else {
+        my $epoch = $time->[c_islocal] ?
+          timelocal(@{$time}[c_sec .. c_mon], $time->[c_year]+1900)
+          :
+          timegm(@{$time}[c_sec .. c_mon], $time->[c_year]+1900);
+        $time->[c_epoch] = $epoch;
+        return $epoch;
+    }
 }
 
 sub hms {
@@ -289,30 +351,75 @@ sub datetime {
     return join($seps{T}, $time->date($seps{date}), $time->time($seps{time}));
 }
 
-# taken from Time::JulianDay
+
+
+# Julian Day is always calculated for UT regardless
+# of local time
 sub julian_day {
     my $time = shift;
-    my ($year, $month, $day) = ($time->year, $time->mon, $time->mday);
-    my ($tmp, $secs);
+    # Correct for localtime
+    $time = $time->gmtime( $time->epoch ) if $time->[c_islocal];
 
-    $tmp = $day - 32075
-      + 1461 * ( $year + 4800 - ( 14 - $month ) / 12 )/4
-      + 367 * ( $month - 2 + ( ( 14 - $month ) / 12 ) * 12 ) / 12
-      - 3 * ( ( $year + 4900 - ( 14 - $month ) / 12 ) / 100 ) / 4
-      ;
+    # Calculate the Julian day itself
+    my $jd = $time->_jd( $time->year, $time->mon, $time->mday,
+                        $time->hour, $time->min, $time->sec);
 
-    return $tmp;
+    return $jd;
 }
 
-# Hi Dominus!
+# MJD is defined as JD - 2400000.5 days
 sub mjd {
     return shift->julian_day - 2_400_000.5;
 }
 
+# Internal calculation of Julian date. Needed here so that
+# both tzoffset and mjd/jd methods can share the code
+# Algorithm from Hatcher 1984 (QJRAS 25, 53-55), and
+#  Hughes et al, 1989, MNRAS, 238, 15
+# See: http://adsabs.harvard.edu/cgi-bin/nph-bib_query?bibcode=1989MNRAS.238.1529H&db_key=AST
+# for more details
+
+sub _jd {
+    my $self = shift;
+    my ($y, $m, $d, $h, $n, $s) = @_;
+
+    # Adjust input parameters according to the month
+    $y = ( $m > 2 ? $y : $y - 1);
+    $m = ( $m > 2 ? $m - 3 : $m + 9);
+
+    # Calculate the Julian Date (assuming Julian calendar)
+    my $J = int( 365.25 *( $y + 4712) )
+      + int( (30.6 * $m) + 0.5)
+        + 59
+          + $d
+            - 0.5;
+
+    # Calculate the Gregorian Correction (since we have Gregorian dates)
+    my $G = 38 - int( 0.75 * int(49+($y/100)));
+
+    # Calculate the actual Julian Date
+    my $JD = $J + $G;
+
+    # Modify to include hours/mins/secs in floating portion.
+    return $JD + ($h + ($n + $s / 60) / 60) / 24;
+}
+
 sub week {
-    # taken from the Calendar FAQ
+    my $self = shift;
+
+    my $J  = $self->julian_day;
+    # Julian day is independent of time zone so add on tzoffset
+    # if we are using local time here since we want the week day
+    # to reflect the local time rather than UTC
+    $J += ($self->tzoffset/(24*3600)) if $self->[c_islocal];
+
+    # Now that we have the Julian day including fractions
+    # convert it to an integer Julian Day Number using nearest
+    # int (since the day changes at midday we convert all Julian
+    # dates to following midnight).
+    $J = int($J+0.5);
+
     use integer;
-    my $J  = shift->julian_day;
     my $d4 = ((($J + 31741 - ($J % 7)) % 146097) % 36524) % 1461;
     my $L  = $d4 / 1460;
     my $d1 = (($d4 - $L) % 365) + $L;
@@ -342,7 +449,16 @@ sub month_last_day {
 
 sub strftime {
     my $time = shift;
-    my $format = @_ ? shift(@_) : "%a, %d %b %Y %H:%M:%S %Z";
+    my $tzname = $time->[c_islocal] ? '%Z' : 'UTC';
+    my $format = @_ ? shift(@_) : "%a, %d %b %Y %H:%M:%S $tzname";
+    if (!defined $time->[c_wday]) {
+        if ($time->[c_islocal]) {
+            return _strftime($format, CORE::localtime($time->epoch));
+        }
+        else {
+            return _strftime($format, CORE::gmtime($time->epoch));
+        }
+    }
     return _strftime($format, (@$time)[c_sec..c_isdst]);
 }
 
@@ -350,8 +466,9 @@ sub strptime {
     my $time = shift;
     my $string = shift;
     my $format = @_ ? shift(@_) : "%a, %d %b %Y %H:%M:%S %Z";
-    my $parsed = _strptime($string, $format);
-    return $time->new($parsed);
+    my @vals = _strptime($string, $format);
+#    warn(sprintf("got vals: %d-%d-%d %d:%d:%d\n", reverse(@vals)));
+    return scalar $time->_mktime(\@vals, (ref($time) ? $time->[c_islocal] : 0));
 }
 
 sub day_list {
@@ -397,10 +514,10 @@ use overload '""' => \&cdate,
 sub cdate {
     my $time = shift;
     if ($time->[c_islocal]) {
-        return scalar(CORE::localtime($time->[c_epoch]));
+        return scalar(CORE::localtime($time->epoch));
     }
     else {
-        return scalar(CORE::gmtime($time->[c_epoch]));
+        return scalar(CORE::gmtime($time->epoch));
     }
 }
 
@@ -419,23 +536,39 @@ use overload
 sub subtract {
     my $time = shift;
     my $rhs = shift;
-    die "Can't subtract a date from something!" if shift;
+    if (UNIVERSAL::isa($rhs, 'Time::Seconds')) {
+        $rhs = $rhs->seconds;
+    }
+
+    if (shift)
+    {
+	# SWAPED is set (so someone tried an expression like NOTDATE - DATE).
+	# Imitate Perl's standard behavior and return the result as if the
+	# string $time resolves to was subtracted from NOTDATE.  This way,
+	# classes which override this one and which have a stringify function
+	# that resolves to something that looks more like a number don't need
+	# to override this function.
+	return $rhs - "$time";
+    }
     
-    if (ref($rhs) && $rhs->isa('Time::Piece')) {
-        return Time::Seconds->new($time->[c_epoch] - $rhs->epoch);
+    if (UNIVERSAL::isa($rhs, 'Time::Piece')) {
+        return Time::Seconds->new($time->epoch - $rhs->epoch);
     }
     else {
         # rhs is seconds.
-        return _mktime(($time->[c_epoch] - $rhs), $time->[c_islocal]);
+        return $time->_mktime(($time->epoch - $rhs), $time->[c_islocal]);
     }
 }
 
 sub add {
     my $time = shift;
     my $rhs = shift;
+    if (UNIVERSAL::isa($rhs, 'Time::Seconds')) {
+        $rhs = $rhs->seconds;
+    }
     croak "Invalid rhs of addition: $rhs" if ref($rhs);
 
-    return _mktime(($time->[c_epoch] + $rhs), $time->[c_islocal]);
+    return $time->_mktime(($time->epoch + $rhs), $time->[c_islocal]);
 }
 
 use overload
@@ -457,6 +590,38 @@ sub compare {
     return $lhs <=> $rhs;
 }
 
+sub add_months {
+    my ($time, $num_months) = @_;
+    
+    croak("add_months requires a number of months") unless defined($num_months);
+    
+    my $final_month = $time->_mon + $num_months;
+    my $num_years = 0;
+    if ($final_month > 11 || $final_month < 0) {
+        # these two ops required because we have no POSIX::floor and don't
+        # want to load POSIX.pm
+        if ($final_month < 0 && $final_month % 12 == 0) {
+            $num_years = int($final_month / 12) + 1;
+        }
+        else {
+            $num_years = int($final_month / 12);
+        }
+        $num_years-- if ($final_month < 0);
+        
+        $final_month = $final_month % 12;
+    }
+    
+    my @vals = _mini_mktime($time->sec, $time->min, $time->hour,
+                            $time->mday, $final_month, $time->year - 1900 + $num_years);
+    # warn(sprintf("got %d vals: %d-%d-%d %d:%d:%d [%d]\n", scalar(@vals), reverse(@vals), $time->[c_islocal]));
+    return scalar $time->_mktime(\@vals, $time->[c_islocal]);
+}
+
+sub add_years {
+    my ($time, $years) = @_;
+    $time->add_months($years * 12);
+}
+
 1;
 __END__
 
@@ -474,7 +639,7 @@ Time::Piece - Object Oriented time objects
 
 =head1 DESCRIPTION
 
-This module replaces the standard localtime and gmtime functions with
+This module replaces the standard C<localtime> and C<gmtime> functions with
 implementations that return objects. It does so in a backwards
 compatible manner, so that using localtime/gmtime in the way documented
 in perlfunc will still return what you expect.
@@ -499,16 +664,18 @@ following methods are available on the object:
     $t->mday                # also available as $t->day_of_month
     $t->mon                 # 1 = January
     $t->_mon                # 0 = January
-    $t->monname             # February
+    $t->monname             # Feb
     $t->month               # same as $t->monname
+    $t->fullmonth           # February
     $t->year                # based at 0 (year 0 AD is, of course 1 BC)
     $t->_year               # year minus 1900
     $t->yy                  # 2 digit year
     $t->wday                # 1 = Sunday
     $t->_wday               # 0 = Sunday
     $t->day_of_week         # 0 = Sunday
-    $t->wdayname            # Tuesday
+    $t->wdayname            # Tue
     $t->day                 # same as wdayname
+    $t->fullday             # Tuesday
     $t->yday                # also available as $t->day_of_year, 0 = Jan 01
     $t->isdst               # also available as $t->daylight_savings
 
@@ -530,7 +697,7 @@ following methods are available on the object:
     $t->tzoffset            # timezone offset in a Time::Seconds object
 
     $t->julian_day          # number of days since Julian period began
-    $t->mjd                 # modified Julian day
+    $t->mjd                 # modified Julian date (JD-2400000.5 days)
 
     $t->week                # week number (ISO 8601)
 
@@ -550,6 +717,11 @@ following methods are available on the object:
                             # see strptime man page. Creates a new
                             # Time::Piece object
 
+Note that C<localtime> and C<gmtime> are not listed above.  If called as
+methods on a Time::Piece object, they act as constructors, returning a new
+Time::Piece object for the current time.  In other words: they're not useful as
+methods.
+
 =head2 Local Locales
 
 Both wdayname (day) and monname (month) allow passing in a list to use
@@ -561,7 +733,7 @@ using locales.
 
   my $french_day = localtime->day(@days);
 
-These settings can be overriden globally too:
+These settings can be overridden globally too:
 
   Time::Piece::day_list(@days);
 
@@ -596,6 +768,20 @@ while examining the object will print the number of seconds (because
 of the overloading), you can also get the number of minutes, hours,
 days, weeks and years in that delta, using the Time::Seconds API.
 
+In addition to adding seconds, there are two APIs for adding months and
+years:
+
+    $t->add_months(6);
+    $t->add_years(5);
+
+The months and years can be negative for subtractions. Note that there
+is some "strange" behaviour when adding and subtracting months at the
+ends of months. Generally when the resulting month is shorter than the
+starting month then the number of overlap days is added. For example
+subtracting a month from 2008-03-31 will not result in 2008-02-31 as this
+is an impossible date. Instead you will get 2008-03-02. This appears to
+be consistent with other date manipulation tools.
+
 =head2 Date Comparisons
 
 Date comparisons are also possible, using the full suite of "<", ">",
@@ -603,10 +789,10 @@ Date comparisons are also possible, using the full suite of "<", ">",
 
 =head2 Date Parsing
 
-Time::Piece links to your C library's strptime() function, allowing
+Time::Piece has a built-in strptime() function (from FreeBSD), allowing
 you incredibly flexible date parsing routines. For example:
 
-  my $t = Time::Piece->strptime("Sun 3rd Nov, 1943",
+  my $t = Time::Piece->strptime("Sunday 3rd Nov, 1943",
                                 "%A %drd %b, %Y");
   
   print $t->strftime("%a, %d %b %Y");
@@ -620,6 +806,8 @@ Outputs:
 For more information see "man strptime", which should be on all unix
 systems.
 
+Alternatively look here: http://www.unix.com/man-page/FreeBSD/3/strftime/
+
 =head2 YYYY-MM-DDThh:mm:ss
 
 The ISO 8601 standard defines the date format to be YYYY-MM-DD, and
@@ -643,12 +831,43 @@ including the ':override' tag in the import list:
 
     use Time::Piece ':override';
 
+=head1 CAVEATS
+
+=head2 Setting $ENV{TZ} in Threads on Win32
+
+Note that when using perl in the default build configuration on Win32
+(specifically, when perl is built with PERL_IMPLICIT_SYS), each perl
+interpreter maintains its own copy of the environment and only the main
+interpreter will update the process environment seen by strftime.
+
+Therefore, if you make changes to $ENV{TZ} from inside a thread other than
+the main thread then those changes will not be seen by strftime if you
+subsequently call that with the %Z formatting code. You must change $ENV{TZ}
+in the main thread to have the desired effect in this case (and you must
+also call _tzset() in the main thread to register the environment change).
+
+Furthermore, remember that this caveat also applies to fork(), which is
+emulated by threads on Win32.
+
+=head2 Use of epoch seconds
+
+This module internally uses the epoch seconds system that is provided via
+the perl C<time()> function and supported by C<gmtime()> and C<localtime()>.
+
+If your perl does not support times larger than C<2^31> seconds then this
+module is likely to fail at processing dates beyond the year 2038. There are
+moves afoot to fix that in perl. Alternatively use 64 bit perl. Or if none
+of those are options, use the L<DateTime> module which has support for years
+well into the future and past.
+
 =head1 AUTHOR
 
 Matt Sergeant, matt@sergeant.org
 Jarkko Hietaniemi, jhi@iki.fi (while creating Time::Piece for core perl)
 
-=head1 License
+=head1 COPYRIGHT AND LICENSE
+
+Copyright 2001, Larry Wall.
 
 This module is free software, you may distribute it under the same terms
 as Perl.
